import numpy as np
import matplotlib.pyplot as plt

# Data reading
x_train = np.array([[1.0,0.5],
                    [0.7,0.7],
                    [1.3,1.0],
                    [1.5,0.3],
                    [1.7,0.9],
                    [2.0,1.5],
                    [1.9,1.3],
                    [2.3,1.5],
                    [2.7,1.7],
                    [3.0,1.3]])
y_train = np.array([1,1,1,1,1,-1,-1,-1,-1,-1])
                     
# Sequential Minimal Optimization 
# Parameters
C = 1.0
N = np.size(x_train,0)

w,b = np.array([0,0]),0

lamb   = np.zeros(N)
lambk  =lamb
lambkp1= np.zeros(N)

# Main iteration
iter = 0
eps  = 1e-10
while iter<1000:    
    for i in range(0,N):
        # Compute w
        sum0,sum1=0,0
        for k in range(0,N):
            sum0=sum0+lambk[k]*y_train[k]*x_train[k,0]
            sum1=sum1+lambk[k]*y_train[k]*x_train[k,1]
        w=np.array([sum0,sum1])
        
        # Ei=f(xi)-yi       
        xi,yi = x_train[i],y_train[i]
        fxi   = np.dot(w,xi)+b
        Ei    = fxi-yi        
        
        # Randomly choose j not equal to i
        j = 0
        while True:
            j = np.random.randint(0,N)
            if j!=i: 
                break        

        # Ej=f(xj)-yj       
        xj,yj = x_train[j],y_train[j]
        fxj   = np.dot(w,xj)+b
        Ej    = fxj-yj
        
        # s=yi*yj
        s     = yi*yj
                
        # Compute L,H
        L,H   = 0,0
        if   s==1:
            L = max(0,lambk[i]+lambk[j]-C)
            H = min(C,lambk[i]+lambk[j]  )
        elif s==-1:
            L = max(0, -lambk[i]+lambk[j])
            H = min(C,C-lambk[i]+lambk[j])
        if L==H:
            continue
        
        # Compute Kii,Kij,Kjj
        Kii = np.dot(xi.T,xi)
        Kij = np.dot(xi.T,xj)
        Kjj = np.dot(xj.T,xj)
        
        if 2*Kij-Kii-Kjj>=0:
            continue
            
        # Update lambdaj
        lambkp1[j] = lambk[j]+yj*(Ej-Ei)/(2*Kij-Kii-Kjj)
        lambkp1[j] = max(lambkp1[j],L)
        lambkp1[j] = min(lambkp1[j],H)
        # Update lambdai
        lambkp1[i] = lambk[i]+s*(lambk[j]-lambkp1[j])
               
        # Update b
        bhat = y_train-w[0]*x_train[:,0]-w[1]*x_train[:,1]
        b    = np.mean(bhat)
                
    print("iter:",iter)
    
    error = np.linalg.norm(lambkp1-lambk)
    
    if error < eps:
        print("Converge!")        
        break
    lambk=lambkp1
    iter = iter+1

print("Training is over!")
x = np.linspace(min(x_train[:,0]),max(x_train[:,0]), 101)
y = -w[0]/w[1]*x - b/w[1]# + 1/omega[1]
plt.plot(x,y)
plt.scatter(x_train[:,0], x_train[:,1])
plt.show()